from __future__ import division
import numpy
import scipy.spatial.distance
import simuPOP 
import simuPOP.utils
import os.path
import matplotlib
import collections
import matplotlib.pyplot as plt

from simuPOP_functions import read_R2_matrix
from generate_genome import expected_r2_WH, expected_r2_SF #, expected_r2_Park, est_c_numerator
from chi2_distribution import generate_bins, find_chi2_of_r2
from realized_Ne import get_matings, get_Ne_from_matings, Hmean_Ne_over_time, variance_Ne_of_subsample, sib_Ne_of_subsample

import logging
import log_config
Results_class_log = logging.getLogger('Results_class')

# FIX SAMPLE SIZE ADJUSTMENT?

class Results():
	def set_genetics(self, directory, name, ext = '.gen', format = 'genepop', adjust = 0):
       	    """
       	    set self.genetics to a simuPOP population from the given genepop file.
       	    only GENEPOP files are supported
       	    """
       	    filename = os.path.join(directory, name + ext)
       	    self.genetics = simuPOP.utils.importPopulation(filename = filename, format = format, adjust = adjust)
	
	def set_r2_plink(self, directory, name, ext = '.plinkR2.npy'):
	    try:
	        self.r2_plink = read_R2_matrix(os.path.join(directory, name + ext))
	    except IOError:
	        ext = ".plinkR2.gz"
	        self.r2_plink = read_R2_matrix(os.path.join(directory, name + ext))
    	    #self.r2_plink is a 1D vector of r2 values
	    return(True)
	   
	def set_subsample(self, base_directory, name, ext = '.subsample'):
	   filename = os.path.join(base_directory, 'samples', 'individuals', name + ext)
	   with open(filename) as INFILE:	           
	       self.subsample = [int(line.strip()) for line in INFILE]
	   return(True)       
		
	def set_whole_genome(self, directory, genome_name, ext = '.genome.tsv.gz.npy'):
	    self.whole_genome = scipy.spatial.distance.squareform(read_R2_matrix(os.path.join(directory, genome_name + ext)))
	    #self.whole_genome is a square distance matrix vector of c values
	    return(True)
	    
	def set_retained_genome(self, name, directory, ext = '.include'):
	   with open(os.path.join(directory, name + ext)) as INFILE:
	       included_loci = [int(line.strip()) for line in INFILE]
	       #select by the rows and columns in the file.
	       self.retained_genome = self.whole_genome[:,included_loci][included_loci,:]
	
	def get_genome_from_name(self, name):
	    name_split = name.split("+")
	    for sec in name_split:
	        if "genome_" in sec:
	            genome_string = sec.split("_")[1]
	    return(genome_string)
	
	def set_genome(self, name, base_directory):
	    genome_name = self.get_genome_from_name(name = name)
	    self.set_whole_genome(genome_name = genome_name, directory = os.path.join(base_directory, 'genomes'))
	    self.set_retained_genome(name = name, directory = os.path.join(base_directory, 'samples', 'loci'))	    
	                 	
	def set_expected_r2(self, name, base_directory, sample_size, popsize, write_WH_SF_exp_r2 = False):
		
		# sample size correction
		ssc = 1/float(sample_size -1)
		
		# Weir and Hill
		r2_WH = expected_r2_WH(c = self.retained_genome, Ne = popsize) + ssc
		Results_class_log.debug("Created WH exp r2 matrix")
		# Sved and Feldman Method
		r2_SF = expected_r2_SF(c = self.retained_genome, Ne = popsize) + ssc
		Results_class_log.debug("Created SF exp r2 matrix")
			
		# Write out expected WH and SF matrices if specified
		if write_WH_SF_exp_r2 is True:
		  #numpy.savetxt(X = r2_WH, fname = os.path.join(base_directory, 'samples', 'expected_r2', name + '.WH_r2_mat.gz'), delimiter = "\t", fmt = '%1.6f')
		  #numpy.savetxt(X = r2_SF, fname = os.path.join(base_directory, 'samples', 'expected_r2', name + '.SF_r2_mat.gz'), delimiter = "\t", fmt = '%1.6f')
		  numpy.save(arr = r2_WH, file = os.path.join(base_directory, 'samples', 'expected_r2', name + '.WH_r2_mat.gz'))
		  numpy.save(arr = r2_SF, file = os.path.join(base_directory, 'samples', 'expected_r2', name + '.SF_r2_mat.gz'))			
			
		  Results_class_log.debug("WH, SF exp r2 files written to path: {}".format((base_directory, 'samples', 'expected_r2', name + "*_r2_mat.gz")))
		
		# Calculate the lower of the two estimates
		lower_matrix = numpy.empty(r2_WH.shape)
		lower_matrix[r2_SF < r2_WH] = r2_SF[r2_SF < r2_WH]
		lower_matrix[r2_WH <= r2_SF] = r2_WH[r2_WH <= r2_SF]
		# write exp r2 to file
		#numpy.savetxt(X = lower_matrix, fname = os.path.join(base_directory, 'samples', 'expected_r2', '{}.lower_r2_mat.gz'.format(name)), delimiter = "\t", fmt = '%1.6f')
		numpy.save(arr = lower_matrix, file = os.path.join(base_directory, 'samples', 'expected_r2', '{}.lower_r2_mat.gz'.format(name)))

		Results_class_log.debug("expected r^2 matrix written to file: {}".format(os.path.join(base_directory, 'samples', 'expected_r2', '{}.lower_r2_mat.gz'.format(name))))
		self.expected_r2 = lower_matrix
		return(True)
			
	def set_expected_r2_unlinked(self, popsize, sample_size):
	    "Includes the effect of sample size"
	    c = .5
            Ne = popsize
            exp_SF = expected_r2_SF(c = c, Ne = Ne, bound = True)
            exp_WH = expected_r2_WH(c = c, Ne = Ne, bound = True)
            expected_r2_unlinked = numpy.min(exp_SF, exp_WH)
            expected_r2_unlinked = expected_r2_unlinked + 1.0/(sample_size - 1)
            self.expected_r2_unlinked = expected_r2_unlinked
            return(True)

	def set_AFS(self, directory, name, ext = '.fold'):
	   """set self.AFS to be a numpy array of MAFs"""
	   AFS = numpy.loadtxt(fname = os.path.join(directory, name + ext))
	   self.AFS = AFS
	
	def set_AFS_frac_lowest_bin(self):
	   self.AFS_frac_lowest_bin = sum(self.AFS < .1) / float(sum(self.AFS > -1))
		
	def set_fraction_c_point5(self):
	   working = scipy.spatial.distance.squareform(self.retained_genome)
	   self.fraction_c_point5 = sum(working == .5) / float(sum(working < 2))

	def set_Ne_estimates(self):
		self.r2_of_subset = dict()
		self.Ne_of_subset = dict()
		self.pairs_of_subset = dict()	
		S = self.sim_params['sample_size']
		# REMOVED Weir SBC 5/1/14
		#Weir_sbc = (float(S)/(S-1))**2\
		#Results_class_log.debug("Sved's sampling bias correction: {}".format(Sved_sbc))
	       
		for subset_key in self.subset_loci_by_c.keys():
		  subset = self.subset_loci_by_c[subset_key]
		  Results_class_log.debug("Subset_shape: {}, r2 shape: {}".format(subset.shape, self.r2_plink.shape))
		  mean_r2 = numpy.mean(self.r2_plink[self.subset_loci_by_c[subset_key]])
		  Ne = self.estimate_Ne_LDNe(mean_r2, S)
		  self.r2_of_subset[subset_key] = mean_r2
		  self.Ne_of_subset[subset_key] = Ne
		  self.pairs_of_subset[subset_key] = int(self.r2_plink[self.subset_loci_by_c[subset_key]].shape[0])
		
	# Subsetting by c
	def subset_by_c(self):
		self.subset_loci_by_c = dict()
		self.set_index_array_c('all', 0, 0.50001)
		self.set_index_array_c('c==0.5', 0.49999, 0.50001)
		#self.set_index_array_c('c>=0.1', 0.1, 0.50001)
		#self.set_index_array_c('c<=0.001', 0, 0.0011)
	
	def set_index_array_c(self, key, c_min, c_max):
	   working = scipy.spatial.distance.squareform(self.retained_genome)
	   indexing_array = numpy.logical_and(working >= c_min, working <= c_max)
	   self.subset_loci_by_c[key] = indexing_array
	
	def set_realized_Ne(self, base_directory):
	   target_dir = os.path.join(base_directory, 'post_coalescent', 'matings')
	   Results_class_log.debug("self.sim_params['current_genome'] = {}".format(self.sim_params['current_genome']))
	   mating_name = "pop_{}+loci_{}+chr_{}+cM_{}+cMunit_{}+ascern_{}+genome_{}+samp_{}+gen_{}+simupop_{}+simcoal_{}+samp_{}.matings".format(
						self.sim_params['popsize'],
						self.sim_params['loci_num'],
						self.sim_params['chr_num'],
						int(self.sim_params['chr_cM']),
						self.sim_params['cM_unit'],
						"01",
						self.sim_params['current_genome'].replace('---', '-'),
						'NA', # sample size
						'NA', # generation
						self.sim_params['simuPOP_rep'],
						self.sim_params['SIMCOAL_rep'],
						'NA'
						)
	   self.matings = get_matings(matings_file = os.path.join(target_dir, mating_name))
	   self.realized_Ne = get_Ne_from_matings(self.matings)
	   return(True)

	# NEW 3-10-14
	def subset_by_r2_max(self, r2_vector, r2_max):
	    #r2_matrix could be ALL or UNLINKED
	    indexing_array = r2_vector <= r2_max
	    return(r2_vector[indexing_array])
	      
	def subset_by_r2_fraction(self, r2_vector, removal_fraction):
	    """removes the highest {removal_fraction}percentage of values from the r2_matrix"""
	    if removal_fraction > 0 and removal_fraction < 1:
	        sorted_vector = numpy.sort(r2_vector)
	        # find number of values to remove (rounded down)
	        to_remove = int(removal_fraction * len(sorted_vector))
	        return(sorted_vector[0:(len(sorted_vector) - to_remove)])
	    else: raise ValueError("removal_fraction should be between 0 and 1")
	    
	def find_best_r2_max(self):
	    corrected_mean = collections.OrderedDict()
	    S = self.sim_params['sample_size']
	    #Weir_sbc = (float(S)/(S-1))**2
	    for xx in numpy.arange(0.01, 1, .01):
	        corrected_mean[xx] = numpy.mean(self.subset_by_r2_max(self.r2_plink, xx)) - 1/(S-1) #* Weir_sbc
	    Results_class_log.debug("corrected_mean from: find_best_r2_max()")   
	    Results_class_log.debug(corrected_mean)
	    # find value in corrected_mean closest to self.expected_2_unlinked
	    # need to ensure that we don't overcorrect.
	    #idx = numpy.abs(numpy.array(corrected_mean.values()) - self.expected_r2_unlinked).argmin()
	    diffs = numpy.array(corrected_mean.values()) - self.expected_r2_unlinked
	    Results_class_log.debug("diffs from: find_best_r2_max()")
	    Results_class_log.debug(diffs)
	    diffs[diffs < 0] = numpy.nan
	    try:
	        idx = numpy.nanargmin(diffs)
	    except ValueError:
	        idx = 0
	    best_correction, corrected_r2 = corrected_mean.items()[idx]
	    return(corrected_r2, best_correction)
	
	def find_best_r2_fraction(self):
	    corrected_mean = collections.OrderedDict()
	    S = self.sim_params['sample_size']
	    #Weir_sbc = (float(S)/(S-1))**2
	    for xx in numpy.arange(0.001, 1, .005):
	        corrected_mean[xx] = numpy.mean(self.subset_by_r2_fraction(self.r2_plink, xx)) - 1/(S-1) #* Weir_sbc
	    Results_class_log.debug("corrected_mean from: find_best_r2_fraction()")   
	    Results_class_log.debug(corrected_mean)
	    # find value in corrected_mean closest to self.expected_2_unlinked
	    #idx = numpy.abs(numpy.array(corrected_mean.values())-self.expected_r2_unlinked).argmin()
	    # need to ensure that we don't overcorrect.
	    diffs = numpy.array(corrected_mean.values()) - self.expected_r2_unlinked
	    Results_class_log.debug("diffs from: find_best_r2_fraction()")
	    Results_class_log.debug(diffs)
	    diffs[diffs < 0] = numpy.nan
	    try:
	        idx = numpy.nanargmin(diffs)
	    except ValueError:
	        idx = 0
	    best_correction, corrected_r2 = corrected_mean.items()[idx]
	    return(corrected_r2, best_correction)

        def chi2_correction(self, expected_count_threshold = 1, delta_r2_threshold = .000001):
            """
            This is a crude way to 'correct' mean r2 based on the chi2 distribution.
            1) Calculate mean r2 from all data
            2) Generate a chi2 distribution using that mean r2
            3) chop off the observed data in bins where the expected value is below the threshold
            4) recalculate r2 and repeat
            """
            # Generate the chi2 dist for the mean r2
            Results_class_log.debug("Starting chi2 correction")
            my_bins = generate_bins(interval = 0.01)
            intervals, counts = find_chi2_of_r2(
                r2 = self.r2_of_subset['all'], num_obs= self.pairs_of_subset['all'], bins = my_bins
                )
            iter_count = 0
            delta_r2 = 1
            working_r2_vector = numpy.copy(self.r2_plink)
            while delta_r2 >= delta_r2_threshold:
                iter_count += 1
                # Find where the expected count goes below threshold
                idx = numpy.nonzero(counts < expected_count_threshold)[0][0]
                # find the threshold bin
                threshold_r2 = intervals[idx]
                old_r2 = numpy.mean(working_r2_vector)
                inital_len = len(working_r2_vector)
                working_r2_vector = self.subset_by_r2_max(r2_vector = working_r2_vector, r2_max = threshold_r2)
                new_r2 = numpy.mean(working_r2_vector)
                delta_r2 = old_r2 - new_r2
                len_diff = inital_len - len(working_r2_vector) 
                Results_class_log.debug("Iter: {}, delta_r2: {}, old_r2: {}, new_r2: {}, threshold_r2: {}, cut: {}".format(iter_count, delta_r2, old_r2, new_r2, threshold_r2, len_diff))
                # update counts
                intervals, counts = find_chi2_of_r2(
                    r2 = new_r2, num_obs= len(working_r2_vector), bins = my_bins
                )
            total_len_diff = len(self.r2_plink) - len(working_r2_vector)
            Results_class_log.debug("Finsihed chi2 correction, new_r2: {}, total_len_diff: {}".format(new_r2, total_len_diff))
            return(new_r2, total_len_diff)


        def weighted_chi2_correction(self, cutoff ):
            my_bins = generate_bins(interval = 0.01)
            exp_intervals, exp_counts = find_chi2_of_r2(
                r2 = self.r2_of_subset['all'], num_obs= self.pairs_of_subset['all'], bins = my_bins
                )       # This is *NOT* using the r2 values after weirs SBC    
                
            # make sure this is inline with the intervals above
            #plt.figure()
            plt.ioff()
            obs_counts, obs_intervals = plt.hist(self.r2_plink, log = True, bins = numpy.arange(0, 1.01, 0.01))[0:2] 
            #plt.figure()
            #plots the ratio of obs to expected counts by r2 value
            # not sure what to do with this yet
            #plt.plot(exp_intervals[exp_intervals < 1] + 0.005, obs_counts / exp_counts[exp_intervals < 1])
            #plt.semilogy()
            #print(zip(
            #    [round(xx, 2) for xx in exp_intervals[exp_intervals < 1] + 0.005], 
            #    [round(xx, 2) for xx in obs_counts / exp_counts[exp_intervals < 1]]
            #    ))
            threshold_r2 = 1
            for idx, jj in enumerate(obs_counts / exp_counts[exp_intervals < 1]):
                if jj > cutoff:
                    threshold_r2 = (exp_intervals[exp_intervals < 1] + 0.005)[idx]
                    break
                
            working_r2_vector = numpy.copy(self.r2_plink)
            working_r2_vector = self.subset_by_r2_max(r2_vector = working_r2_vector, r2_max = threshold_r2)
            new_r2 = numpy.mean(working_r2_vector)
            return(new_r2, threshold_r2)
            
	# Analysis
	def estimate_Ne_LDNe(self, mean_r2, S):
		# edited 9/10/15
		mean_r2 = mean_r2 * (S/(S-1))**2
		#r2_sample = 1.0/(S-1) # Sved 2013 PLOS
		r2_sample = (1.0/S) + 3.19/(S*S)
		r2_prime = mean_r2 - r2_sample
		if 1/9.0 - 2.76*r2_prime > 0:
			numerator = 1/3.0 + numpy.sqrt(1/9.0 - 2.76*r2_prime)
			denominator = 2*r2_prime
			Ne = (numerator/denominator)
		else:
			Ne = (1/3.0) * (2* r2_prime)
		return(Ne)
	
	def r2_difference(self, obs_r2, exp_r2, subset_key = None):
		if subset_key is None:
			return(obs_r2 - exp_r2)
		else:
			return(obs_r2[self.subset_loci_by_c[subset_key]] - exp_r2[self.subset_loci_by_c[subset_key]])
			
	def plot_r2_difference(self, subset_key = None):			
		obs_r2 = self.r2_plink
		exp_r2 = self.expected_r2
		r2_diff = self.r2_difference(obs_r2 = obs_r2, exp_r2 = exp_r2, subset_key = subset_key)
		matplotlib.pyplot.hist(r2_diff, 
			log = True, 
			bins = 100, 
			range = (-1,1)
			)
		matplotlib.pylab.title("Mean Difference: {}\n{} Subset = {}".format(numpy.mean(r2_diff), self.name, subset_key))
		matplotlib.pylab.xlabel('R2 diffence')
		matplotlib.pylab.ylabel('Count')
	
	def plot_r2_distribution(self, subset_key = None):
		obs_r2 = self.r2_plink
		exp_r2 = self.expected_r2
		if subset_key is None:
			pass
		else:
			obs_r2 = obs_r2[self.subset_loci_by_c[subset_key]]
			exp_r2 = exp_r2[self.subset_loci_by_c[subset_key]]
		matplotlib.pyplot.hist(obs_r2, log = True, bins = 50, range = (0,1), alpha = .5, color = "Blue")
		matplotlib.pyplot.hist(exp_r2, log = True, bins = 50, range = (0,1), alpha =.5, color = "Green")
		matplotlib.pylab.xlim(0, 1)
		matplotlib.pylab.ylim(ymin = 1)
		matplotlib.pylab.title("Blue = Observed / Green = Expected\n{}, Subset = {}".format(self.name, subset_key))
		matplotlib.pylab.ylabel("Count")
		matplotlib.pylab.xlabel("R2")
		
	# Observed vs Expected R2
	def plot_obs_exp_r2(self, subset_key = None):
		obs_r2 = self.r2_plink
		exp_r2 = self.expected_r2
		if subset_key is None:
			pass
		else:
			obs_r2 = obs_r2[self.subset_loci_by_c[subset_key]]
			exp_r2 = exp_r2[self.subset_loci_by_c[subset_key]]
			
		matplotlib.pyplot.hexbin(exp_r2, obs_r2,
			mincnt = 1, bins = 'log', gridsize = 30)
		matplotlib.pylab.ylim([-0.04, 1.04])
		matplotlib.pylab.xlim([-0.04, 1.04])
		matplotlib.pylab.title("exp R2 vs OBS R2\n{}, Subset = {}".format(self.name, subset_key))
		matplotlib.pylab.xlabel('Lower of [SF, WH] expected R2 ')
		matplotlib.pylab.ylabel('Observed R2')
		cb = matplotlib.pyplot.colorbar()
		cb.set_label('log(count)')
		matplotlib.pyplot.show()
	
	# Observed R2 vs Recombination Fraction
	def plot_c_obs_r2(self, subset_key = None, xscale = 'linear'):
		c_mat = scipy.spatial.distance.squareform(self.retained_genome)
		obs_r2 = self.r2_plink
		if subset_key is None:
			pass
		else:
			c_mat = c_mat[self.subset_loci_by_c[subset_key]]
			obs_r2 = obs_r2[self.subset_loci_by_c[subset_key]]
		matplotlib.pyplot.hexbin(c_mat, obs_r2,
			mincnt = 1, bins = 'log', gridsize = 60,
				xscale = xscale
			 )
		matplotlib.pylab.ylim([-0.04, 1.04])
		matplotlib.pylab.xlim([-0.02, 0.52])
		matplotlib.pylab.title("c vs Observed R2\n{}, Subset = {}".format(self.name, subset_key))
		matplotlib.pylab.xlabel('Recombination Fraction (c)')
		matplotlib.pylab.ylabel('Observed R2')
		cb = matplotlib.pyplot.colorbar()
		cb.set_label('log(count)')
		matplotlib.pyplot.show()
		
	def plot_r2_hist(self, subset_key = None):
		obs_r2 = self.r2_plink
		if subset_key is None:
			pass
		else:
			obs_r2 = obs_r2[self.subset_loci_by_c[subset_key]]
		matplotlib.pyplot.hist(obs_r2, log = True)
			

class Sim_Results(Results):
	def __init__(self, directory, name, generate_expected_r2 = False):
	        Results_class_log.debug("Sim_Results.__init__(self, directory = {}, name = {}, generate_expected_r2  = {})".format(directory, name, generate_expected_r2))
		self.name = name
		self.directory = directory
		self.set_sim_params()
		self.set_genetics(directory = os.path.join(self.directory, 'samples', 'genepop'), name = self.name)
		self.set_r2_plink(directory = os.path.join(self.directory, 'samples', 'plink', 'output'), name = self.name)
		self.set_genome(base_directory = self.directory, name = self.name)
		if generate_expected_r2:
		  #self.set_expected_r2(directory = os.path.join(self.directory, 'samples', 'expected_r2'), name = self.name)
		  self.set_expected_r2(name = self.name, base_directory = self.directory, 
						 sample_size = self.sim_params['sample_size'], popsize = self.sim_params['popsize'],
						 write_WH_SF_exp_r2 = False)
		self.set_expected_r2_unlinked(popsize = self.sim_params['popsize'], sample_size = self.sim_params['sample_size'])
		self.set_AFS(directory = os.path.join(self.directory, 'samples', 'fold'), name = self.name)
		self.subset_by_c()
		self.set_AFS_frac_lowest_bin()
		self.set_fraction_c_point5()
		self.set_Ne_estimates()
		self.set_realized_Ne(base_directory = self.directory)
		self.set_subsample(base_directory = self.directory, name = self.name)
		
	def set_sim_params(self):
		"""Gets simulation parameters from the file name"""
		#popsize, sample_size, generation, loci_num, chr_num, chr_cM, scale_recombination, scale_mutation, rep = self.name.split("_")
		first_splits = self.name.split("+")
		popsize, loci_num, chr_num, chr_cM, cM_unit, ascertainment_AF, current_genome, sample_size, generation, simuPOP_rep, SIMCOAL_rep, sampling_rep  = ((xx.split("_"))[1] for xx in first_splits)
		self.sim_params = dict()
		self.sim_params['popsize'] = int(popsize)
		self.sim_params['loci_num'] = int(loci_num)
		self.sim_params['chr_num'] = int(chr_num)
		self.sim_params['chr_cM'] = float(chr_cM)
		self.sim_params['cM_unit'] = int(cM_unit)
		self.sim_params['ascertainment_AF'] = float(ascertainment_AF)
		#self.sim_params['current_genome'] = "-".join([str(jj) for jj in current_genome])
		self.sim_params['current_genome'] = str(current_genome)
		self.sim_params['sample_size'] = int(sample_size)
		self.sim_params['generation'] = int(generation)
		self.sim_params['simuPOP_rep'] = int(simuPOP_rep)
		self.sim_params['SIMCOAL_rep'] = int(SIMCOAL_rep)
		self.sim_params['sampling_rep'] = int(sampling_rep)
		return(True)

	def summary(self, *args):
	        Results_class_log.debug("Start summary")
		summary = collections.OrderedDict()
		#Simulation parameters
		summary['popsize'] =  self.sim_params['popsize']
		summary['starting_loci'] =  self.sim_params['loci_num']
		summary['chr_num'] =  self.sim_params['chr_num']
		summary['chr_cM'] =  self.sim_params['chr_cM']
		summary['cM_unit'] =  self.sim_params['cM_unit']
		summary['ascertainment_AF'] = self.sim_params['ascertainment_AF']
		summary['sample_size'] =  self.sim_params['sample_size']
		summary['generation'] =  self.sim_params['generation']
		summary['current_genome'] =  self.sim_params['current_genome']
		summary['SIMCOAL_rep'] =  self.sim_params['SIMCOAL_rep']
		summary['simuPOP_rep'] =  self.sim_params['simuPOP_rep']
		summary['sampling_rep'] =  self.sim_params['sampling_rep']
		
		#Final Loci
		summary['final_loci '] = len(self.AFS)
		
		#AFS
		summary['AFS_lowest_bin'] =  self.AFS_frac_lowest_bin
		
		# Recombination
		summary['fraction_c_point5'] =  self.fraction_c_point5
		summary['min_c'] =  numpy.min(scipy.spatial.distance.squareform(self.retained_genome))
		
		#r^2
		summary['expected_r2_unlinked'] = self.expected_r2_unlinked

		summary['mean_r2_all']   = self.r2_of_subset['all']
		summary['mean_r2_unlinked']  = self.r2_of_subset['c==0.5']
		
		# Ne 
		summary['Ne_all']   = self.Ne_of_subset['all']
		#summary['Ne_almost_unlinked']  = self.Ne_of_subset['c>=0.1']
		summary['Ne_unlinked']  = self.Ne_of_subset['c==0.5']
		#summary['Ne_closely_linked']  = self.Ne_of_subset['c<=0.001']
		
		# pairs of subset
		summary['pairs_all']  = self.pairs_of_subset['all']
		#summary['pairs_almost_unlinked']  = self.pairs_of_subset['c>=0.1']
		summary['pairs_unlinked']  = self.pairs_of_subset['c==0.5']
		#summary['pairs_closely_linked']  = self.pairs_of_subset['c<=0.001']
		
		# realized Ne
		summary['one_generation_realized_Ne'] = Hmean_Ne_over_time(self.realized_Ne, generations = [9])
		summary['ten_generation_realized_Ne'] = Hmean_Ne_over_time(self.realized_Ne, generations = range(9))
		summary['subsample_realized_Ne'] = variance_Ne_of_subsample(subsample = self.subsample, matings = self.matings, generation = 9)
		summary['sib_Ne_of_subsample'] = sib_Ne_of_subsample(subsample = self.subsample, matings = self.matings, generation = 9)
		
		# attempted at r2 corrections
		Results_class_log.debug("Start r^2 corrections")
		
		# r2 based
#		summary['best_r2_max_correction_r2'], summary['best_r2_max_correction_threshold'] = self.find_best_r2_max()
#		summary['best_r2_max_correction_Ne'] = self.estimate_Ne_LDNe( summary['best_r2_max_correction_r2'], S = self.sim_params['sample_size'])
#
#		
#		summary['best_r2_fraction_correction_cutoff_r2'], summary['best_r2_fraction_correction_cutoff_threshold'] = self.find_best_r2_fraction()
#		summary['best_r2_fraction_correction_cutoff_Ne'] = self.estimate_Ne_LDNe(summary['best_r2_fraction_correction_cutoff_r2'], S = self.sim_params['sample_size'])
#		
#		summary['chi2_correction_r2_cutoff_3'], summary['chi2_correction_threshold_3'] = self.weighted_chi2_correction(cutoff = 3)
#		summary['chi2_correction_Ne_cutoff_3'] = self.estimate_Ne_LDNe(summary['chi2_correction_r2_cutoff_3'], S = self.sim_params['sample_size'])
#		
#		summary['chi2_correction_r2_cutoff_10'], summary['chi2_correction_threshold_10'] = self.weighted_chi2_correction(cutoff = 10)
#		summary['chi2_correction_Ne_cutoff_10'] = self.estimate_Ne_LDNe(summary['chi2_correction_r2_cutoff_10'], S = self.sim_params['sample_size'])		

		Results_class_log.debug("Finish summary chi2_correction()")
		return(summary)
		