from __future__ import division

import simuPOP 
import simuPOP.utils
import simuPOP.sampling
import sys
import os
import os.path
import itertools
import subprocess
import shlex
import glob
import re

from Genome_class import Genome
from PGD_spider import pgd_arlequin_to_genepop
from simuPOP_functions import process_subsample, write_parents #, filter_MAF, get_folded_freqs
from PARAMS import BASE_DIR, SIMCOAL_EXE, SIMCOAL_FORMAT, PDG_SPIDER_EXE, SPID #, PLINK_EXE, DEBUG

from simulation_settings import  fixed_families

if fixed_families:
    from controlled_mating import fixedChooser #, parents_of_N

import logging
Sim_Batch_Class_log = logging.getLogger('Sim_Batch_Class')

class Sim_Batch:
	def __init__(self,
				name,
				# Constants
				ploidy 		= 2,
				map_function 	= "kosambi",
				mutation_rate 	= 0,
				p_crit  	= 0.05,
				fixed_families  = False, 
				
				# Simulation Variables
				popsizes 		= [],
				sampsizes 		= [],
				generations 	= [],
				SIMCOAL_min_allele_freqs = [.1],
				
				# Genome Variable
				loci_num		= [],
				chr_num 		= [],
				chr_cM 			= [],
				cM_unit 		= [],
				
				#Replication
				GENOME_replicates 	= None,
				SIMCOAL_replicates	= None,
				simuPOP_replicates 	= None,
				sampling_replicates     = None
				):
				
		# constants		
		self.name 		= name
		self.ploidy 		= ploidy 		
		self.map_function 	= map_function
		self.mutation_rate 	= mutation_rate
		self.p_crit 		= p_crit
		self.fixed_families = fixed_families
		
		# Replication
		self.GENOME_replicates = GENOME_replicates
		self.SIMCOAL_replicates = SIMCOAL_replicates
		self.simuPOP_replicates = simuPOP_replicates
		self.sampling_replicates = sampling_replicates
		
		# Simulation Variables		
		self.popsizes 				= popsizes
		self.sampsizes 				= sampsizes
		self.generations			= generations
		self.SIMCOAL_min_allele_freqs           = SIMCOAL_min_allele_freqs
		
		# Genome Variables
		self.loci_num				= loci_num
		self.chr_num 				= chr_num
		self.chr_cM 				= chr_cM
		self.cM_unit				= cM_unit
		
		# Validation
		for ii in self.loci_num:
			for jj in self.chr_num:
				if ii % jj != 0:
					raise ValueError("Please ensure that loci_num can be evenly be divided across chr_num")

		# Ensure replication values are integer-like
		for ii in [self.GENOME_replicates, self.SIMCOAL_replicates, self.simuPOP_replicates, self.sampling_replicates ]:
		  try:
		      int(ii)
		  except TypeError:
		      raise
		  
	
		self.params = self.set_genome_combinations()
		#if DEBUG: print("Finish construct_paramaters()")
		#if DEBUG: print("Finish Sim_Batch.__init__()")
		
	def set_genome_combinations(self):
		genome_combinations = itertools.product(self.loci_num,
							self.chr_num,
							self.chr_cM,
							self.cM_unit,
							#self.SIMCOAL_min_allele_freqs
							)
		param_keys = ['loci_num', 'chr_num', 'chr_cM', 'cM_unit']
		self.genome_combinations = [dict(zip(param_keys, combination)) for combination in genome_combinations]
		return(True)
		
	def run_sims(self, path = BASE_DIR):
		#For each possible genome
		genome_combination_counter = 1
		with open(os.path.join(path, 'genomes', 'genome_parameters.tsv'), 'w') as OUTFILE:
		    OUTFILE.write("{}\t".format('Genome'))
		    OUTFILE.write("\t".join(['loci_num', 'chr_num', 'chr_cM', 'cM_unit']))
		    OUTFILE.write("\n")
		    for genome_param_set in self.genome_combinations:
		        OUTFILE.write("{}\t".format(genome_combination_counter))
		        OUTFILE.write("\t".join([str(jj) for jj in genome_param_set.values()]))
		        OUTFILE.write("\n")
		        genome_combination_counter += 1
		
		genome_combination_counter = 0
		for genome_param_set in self.genome_combinations:
		  genome_combination_counter += 1
		  Sim_Batch_Class_log.debug("Start Genome Combination: {}".format(genome_combination_counter))
		  Sim_Batch_Class_log.debug("{}".format(genome_param_set.items()))
		  # Genome Replication
		  for genome_rep in range(self.GENOME_replicates):
				genome_rep +=1 # match the 1-based indexing
		
				#if DEBUG: print("Creating Genome instance: {}".format(genome_rep))
				Sim_Batch_Class_log.debug("Creating Genome instance: {}".format(genome_rep))
				# Create Genome
				sim_Genome = Genome(chr_num = genome_param_set['chr_num'], 
					chr_cM = genome_param_set['chr_cM'],
					chr_loci = int(genome_param_set['loci_num']/genome_param_set['chr_num']),
					cM_unit = genome_param_set['cM_unit'],
					map_function = self.map_function, 
					)
					
				current_genome = [genome_combination_counter, genome_rep]
				
				#if DEBUG: print("Finished creating Genome instance, replicate: {}".format(genome_rep))				
				Sim_Batch_Class_log.debug("Finished creating Genome instance, replicate: {}".format(genome_rep))
				
				# WRITE full genome c_matrix OUT HERE
				sim_Genome.write_c_matrix_to_file(os.path.join(path, 'genomes'), '{}-{}.genome'.format(genome_combination_counter, genome_rep))
				
				# SIMCOAL 
				# Loop over combinations of population sizes and ascertainment bias 
				for popsize, min_allele_freq in itertools.product(self.popsizes, self.SIMCOAL_min_allele_freqs):
					SIMCOAL_name = "pop_{}+loci_{}+chr_{}+cM_{}+cMunit_{}+ascern_{}+genome_{}+samp_{}+gen_{}+simupop_{}+simcoal".format(
						popsize,
						genome_param_set['loci_num'],
						genome_param_set['chr_num'],
						genome_param_set['chr_cM'],
						genome_param_set['cM_unit'],
						#str(genome_param_set['ascertainment_AF']).replace('.', ''),
						str(min_allele_freq).replace('.', ''),
						"-".join([str(jj) for jj in current_genome]), # genome_rep
						'NA', # sample size
						'NA', # generation
						'NA' # simupop_rep
						# simcoal_rep
						)
					SIMCOAL_infile = os.path.join(path, 'coalescent', "{}.par".format(SIMCOAL_name))
					Sim_Batch_Class_log.debug("Creating SIMCOAL infile: {}".format(SIMCOAL_infile))
					self.construct_SIMCOAL_file(file_name = SIMCOAL_infile,	Genome = sim_Genome, popsize = popsize, min_allele_freq = min_allele_freq)
					Sim_Batch_Class_log.debug("Start run_SIMCOAL()")
					#self.run_SIMCOAL(in_file = "{}.par".format(SIMCOAL_name), reps = self.SIMCOAL_replicates, SIMCOAL_exec_path = SIMCOAL_EXE, simcoal_format = SIMCOAL_FORMAT)
					self.run_SIMCOAL(in_file = "{}.par".format(SIMCOAL_name), reps = self.SIMCOAL_replicates, SIMCOAL_exec_path = SIMCOAL_EXE, simcoal_format = SIMCOAL_FORMAT)
					Sim_Batch_Class_log.debug("Finish run_SIMCOAL()")
					
					# Move this to a separate function!
					if SIMCOAL_FORMAT == "FASTSIMCOAL2":
					   Sim_Batch_Class_log.debug("Start replacement of 'DataType=DNA' with 'DataType=SNP' within SIMCOAL .arp files")
					   origWD = os.getcwd() # remember our original working directory
					   os.chdir(os.path.join(os.path.abspath(sys.path[0]), BASE_DIR))
					   files_to_reword = glob.glob(os.path.join('.', 'coalescent', "{}".format(SIMCOAL_name), "*.arp" ))
					   for my_file in files_to_reword:
					       with open(my_file) as INFILE:
					           my_lines = INFILE.readlines()
					       with open(my_file, 'w') as OUTFILE:
					           line_iter = iter(my_lines)
					           for line in line_iter:
					               if 'DataType=DNA' in line:
					                   OUTFILE.write(line.replace('DataType=DNA', 'DataType=SNP'))
					               elif line.startswith("1_"):
					                   sample, hap1, haplotype = line.split('\t')
					                   translated_haplotype = haplotype.replace('0', '3')
					                   first_line = "\t".join([sample, hap1, translated_haplotype])
					                   second_line = next(line_iter).replace('0', '3')
					                   OUTFILE.write(first_line)
					                   OUTFILE.write(second_line)              
					               else:
					                   OUTFILE.write(line)
					   os.chdir(origWD)
					   Sim_Batch_Class_log.debug("Finished replacement of 'DataType=DNA' with 'DataType=SNP' within SIMCOAL .arp files.  Searched: {} files".format(len(files_to_reword)))
					
					# Rename SIMCOAL64 files to match new fastsimcoal2  format
					# Move this to a separate function!
					if SIMCOAL_FORMAT == "SIMCOAL2_64":
					    Sim_Batch_Class_log.debug("Start rename of SIMCOAL2_64 files.")
					    origWD = os.getcwd()
					    os.chdir(os.path.join(os.path.abspath(sys.path[0]), BASE_DIR))
					    files_to_rename = glob.glob(os.path.join('.', 'coalescent', "{}".format(SIMCOAL_name), "*.arp" ))
					    temp_idx = 1
					    for my_file in files_to_rename:
					        if "0.arp" in my_file: #matches old naming scheme
					            new_name = re.sub(pattern = '_[1-9]*0.arp', repl = "_1_{}.arp".format(temp_idx), string = my_file)
					            os.rename(my_file, new_name)
					            Sim_Batch_Class_log.debug("Renamed {} to {}.".format(my_file, new_name))
					            temp_idx +=1				            

					   
					#Convert Each SIMCOAL replicate to genepop
					#Set chromosome names - ensure they are outside the range of human chromosomes (for PLINK)
					CHROM_NAMES = [str(100 + ii) for ii in range(genome_param_set['chr_num'])]
					LOCUS_NAMES = [str(ii) for ii in range(genome_param_set['loci_num'])]
					#if DEBUG: print("Start conversion to Genepop")
					Sim_Batch_Class_log.debug("Start conversion to Genepop")
					for SIMCOAL_rep in range(self.SIMCOAL_replicates):
						PGD_infile = os.path.join(BASE_DIR, 'coalescent', SIMCOAL_name, SIMCOAL_name + "_1_{}.arp".format(SIMCOAL_rep +1 ))
						#if DEBUG: print("PGD_infile: {}".format(PGD_infile))
						Sim_Batch_Class_log.debug("PGD_infile: {}".format(PGD_infile))
						genepop0_file = pgd_arlequin_to_genepop(spid = SPID, pgd_exe = PDG_SPIDER_EXE, in_file = PGD_infile, output_dir = os.path.join(BASE_DIR, 'post_coalescent') )
						for simuPOP_rep in range(self.simuPOP_replicates):
							my_pop = simuPOP.Population(
								size = popsize, # number of individuals
								ploidy = self.ploidy, 
								loci = [int(genome_param_set['loci_num']/genome_param_set['chr_num'])] * genome_param_set['chr_num'], # number of loci per chromosome * number of chromosomes
								alleleNames = ('C', 'A', 'G', 'T'), # awkward conversions as PGD spider
								lociPos = list(itertools.chain(sim_Genome.simuPOP_lociPos)),# positions of the loci given cumulative, adjacent recombination fractions
								lociNames = LOCUS_NAMES,
								chromNames = CHROM_NAMES,
								infoFields=['father_idx', 'mother_idx']
								)
							self.set_genotypes_from_genepop(pop = my_pop, genepop_file = genepop0_file)
							simuPOP_dict = {'popsize': popsize,
											'loci_num': genome_param_set['loci_num'],
											'chr_num': genome_param_set['chr_num'],
											'chr_cM' : genome_param_set['chr_cM'],
											'samp_sizes' : self.sampsizes,
											'p_crit' : self.p_crit,
											'rep' : simuPOP_rep,
											'SIMCOAL_rep' : SIMCOAL_rep,
											'cM_unit' : genome_param_set['cM_unit'],
											'ascertainment_AF' : min_allele_freq,
											'current_genome': current_genome, 
											'sampling_replicates': self.sampling_replicates
												}
							self.evolve_forward(pop = my_pop, Genome = sim_Genome, param_dict = simuPOP_dict, fixed_mating = self.fixed_families )
					#if DEBUG: print("Finish.")
					Sim_Batch_Class_log.debug("End Simualation Batch")

	
	def evolve_forward(self, pop, Genome, param_dict, fixed_mating = False):
		Sim_Batch_Class_log.debug("Starting Forward Simulation. Sampling at generations: {}".format(self.generations))
		
		# determine mating scheme
		if fixed_mating is True:
			my_matingScheme = simuPOP.HomoMating(
				chooser = simuPOP.PyParentsChooser(fixedChooser),
				generator = simuPOP.OffspringGenerator(
				    sexMode=(simuPOP.GLOBAL_SEQUENCE_OF_SEX, simuPOP.MALE, simuPOP.FEMALE),
				    ops = [simuPOP.Recombinator(intensity = 1),simuPOP.ParentsTagger()]),
				subPopSize = pop.popSize()
			)
		elif fixed_mating is False:
			my_matingScheme = simuPOP.RandomMating(
				ops = [simuPOP.Recombinator(intensity = 1), simuPOP.ParentsTagger()],
				sexMode = (simuPOP.GLOBAL_SEQUENCE_OF_SEX, simuPOP.MALE, simuPOP.FEMALE)
			)
		# construct postOps_dict
		postOps_dict = {
			'popsize': param_dict['popsize'],
			'loci_num': param_dict['loci_num'],
			'chr_num': param_dict['chr_num'],
			'chr_cM' : param_dict['chr_cM'],
			'samp_sizes' : self.sampsizes,
			'Genome': Genome,
			'write_WH_SF_exp_r2': False,
			'write_genepop' : True,
			'write_map_ped' : True,
			'path' : BASE_DIR,
			'ascertainment_AF' : param_dict['ascertainment_AF'],
			'popsize': param_dict['popsize'],
			'loci_num': param_dict['loci_num'],
			'chr_num': param_dict['chr_num'],
			'chr_cM' : param_dict['chr_cM'],
			'rep' : param_dict['rep'],
			'SIMCOAL_rep' : param_dict['SIMCOAL_rep'],
			'cM_unit' : param_dict['cM_unit'],
			'current_genome' : param_dict['current_genome'],
			'sampling_replicates': self.sampling_replicates
		}
		# evolve the population
		pop.evolve(
			initOps = [simuPOP.InitSex(maleProp = .5)],
			matingScheme = my_matingScheme,	
			preOps = [],
			postOps= [
				simuPOP.PyOperator(func = write_parents, param = postOps_dict),
				simuPOP.PyOperator(func = process_subsample, param = postOps_dict, at = self.generations)
			],
			finalOps = [],
			gen = max(self.generations) + 1
		)

	def set_genotypes_from_genepop(self, pop, genepop_file, adjust = -1):
		with open (genepop_file, 'r') as INFILE:
			line_count = 0
			found_first_pop = False
			locus_names = list()
			individuals = list()
			genotypes = list()
			for line in INFILE:
				line_count += 1
				if line_count == 1:
					header = line.strip()
					header = header
				else:
					if not(found_first_pop):
						if line.strip().lower() == 'pop':
							found_first_pop = True
							continue
						else:
							locus_name = line.strip()
							locus_names.append(locus_name)
							continue
					else:
						ind, geno_line = line.strip().split(',')
						individuals.append(ind) # genopop individuals
						genotypes.append(geno_line)
			for ind, genos in zip(pop.individuals(), genotypes):
				chr_0 = list()
				chr_1 = list()
				geno_split = genos.split()
				for geno in geno_split:
					if len(geno) == 6:
						first = int(geno[2:3]) + adjust
						second = int(geno[5:6]) + adjust
					elif len(geno) == 4:
						first = int(geno[1:2]) + adjust
						second = int(geno[3:4]) + adjust
					else:
						raise ValueError("Genotypes should be four or 6 characters")
					chr_0.append(first) 
					chr_1.append(second)
				ind.setGenotype(chr_0, 0)
				ind.setGenotype(chr_1, 1)			
		#return(locus_names, zip(individuals, genotypes))
				
	def construct_SIMCOAL_file(self, file_name, Genome, popsize, min_allele_freq):
		file_lines = list()
		file_lines.append("//Parameters for the coalescence simulation\n")
		file_lines.append("1 samples to simulate\n")
		file_lines.append("//Population effective sizes\n")
		file_lines.append("{}\n".format(2*popsize)) #SIMCOAL works with haploid genomes
		file_lines.append("//Sample sizes\n")
		file_lines.append("{}\n".format(2*popsize)) #SIMCOAL works with haploid genomes
		file_lines.append("//Growth rates\n")
		file_lines.append("0\n")
		file_lines.append("//Number of migration matrices : 0 implies no migration between demes\n")
		file_lines.append("0\n")
		file_lines.append("//historical events\n")
		file_lines.append("0 historical events\n")
		file_lines += Genome.SIMCOAL_genome_strings(min_allele_freq = min_allele_freq)
		with open(file_name, 'w') as OUTFILE:
			OUTFILE.writelines(file_lines)
		
	def run_SIMCOAL(self, in_file, reps, SIMCOAL_exec_path, simcoal_format):
		#in_file = os.path.split(in_file)[1]
		if simcoal_format == "FASTSIMCOAL2":
		    exec_string = '"{}" -i "{}" -n {} -g -p -q'.format(SIMCOAL_exec_path, in_file, reps)
		elif simcoal_format == "SIMCOAL2_64":
		    exec_string = '"{}" {} {} 1'.format(SIMCOAL_exec_path, in_file, reps)
		Sim_Batch_Class_log.debug("SIMCOAL exec_string: {}".format(exec_string))
		origWD = os.getcwd() # remember our original working directory
		#os.chdir(os.path.join(os.path.abspath(sys.path[0]), BASE_DIR, 'coalescent'))
		os.chdir(os.path.join(BASE_DIR, 'coalescent'))
		subprocess.call(shlex.split(exec_string))
		os.chdir(origWD)
		


	
		
		
		
