import numpy
import matplotlib
import pylab

from simuPOP_functions import *

class Sim_Results():
	"""Represents simulation  for one unique combination of parameters.  
	Class is used for plotting or for access to data matrices."""
	def __init__(self, directory, base_name, sim = True):
		self.name = base_name
		if sim is True:
			self.set_sim_params() 
		
		self.r2_plink = read_R2_matrix(os.path.join(directory, base_name + '.plinkR2'))
		self.c_mat = read_R2_matrix(os.path.join(directory, base_name + '.c_mat.gz'))
		self.lower_r2_estimate = read_R2_matrix(os.path.join(directory, base_name + '.lower_r2_mat.gz'))
		
		self.set_AFS(directory, base_name)
		self.set_mean_r2()
		self.final_loci = len(self.AFS)
		#self.set_unlinked_exp_r2()
		
		self.subset_loci = dict()
		self.set_index_array_c(.5, 0.49999, 0.50001)
		self.set_index_array_c(.1, 0.09, 0.11)
		self.set_index_array_c(.01, 0, 0.01)
		
	def set_sim_params(self):
		popsize, sample_size, generation, loci_num, chr_num, chr_cM, scale_recombination, scale_mutation, rep = self.name.split("_")
		self.popsize = int(popsize)
		self.sample_size = int(sample_size)
		self.generation = int(generation)
		self.loci_num = int(loci_num)
		self.chr_num = int(chr_num)
		self.chr_cM = float(chr_cM)
		self.scale_recombination = float(scale_recombination)
		self.scale_mutation = float(scale_mutation)
		self.rep = int(rep)
		
	def set_AFS(self, directory, base_name, ext = '.fold'):
		# Allele Frequency Spectrum
		AFS = numpy.loadtxt(fname = os.path.join(directory, base_name + ext))
		self.AFS = AFS
		self.meanAF = numpy.mean(self.AFS)
		
	def set_mean_r2(self):
		self.mean_obs_r2 = numpy.mean(self.r2_plink)
		self.mean_exp_r2 = numpy.mean(self.lower_r2_estimate)
		
	def set_index_array_c(self, key, c_min, c_max):
		indexing_array = numpy.logical_and(self.c_mat >= c_min, self.c_mat <= c_max)
		self.subset_loci[key] = indexing_array
		
	def set_unlinked_exp_r2(self):
		c = .5
		Ne = self.popsize
		S = self.sample_size
		exp_SF = expected_r2_SF(c = c, Ne = Ne, S = S, bound = True)
		exp_WH = expected_r2_WH(c = c, Ne = Ne, S = S, bound = True)
		self.unlinked_exp_r2 = min(exp_SF, exp_WH)
	
	def r2_difference(self, obs_r2, exp_r2, subset_key = None):
		if subset_key is None:
			return(obs_r2 - exp_r2)
		else:
			return(obs_r2[subset_loci[subset_key]] - exp_r2[subset_loci[subset_key]])
		
	def plot_r2_difference(self, subset_key = None):			
		obs_r2 = self.r2_plink
		exp_r2 = self.lower_r2_estimate
		r2_diff = self.r2_difference(obs_r2 = obs_r2, exp_r2 = exp_r2, subset_key = subset_key)
		matplotlib.pyplot.hist(r2_diff, 
			log = True, 
			bins = 100, 
			range = (-1,1)
			)
		pylab.title("Mean Difference: {}\n{} Subset = {}".format(numpy.mean(r2_diff), self.name, subset_key))
		pylab.xlabel('R2 diffence')
		pylab.ylabel('Count')
	
	def plot_r2_distribution(self, subset_key = None):
		obs_r2 = self.r2_plink
		exp_r2 = self.lower_r2_estimate
		if subset_key is None:
			pass
		else:
			obs_r2 = obs_r2[subset_loci[subset_key]]
			exp_r2 = exp_r2[subset_loci[subset_key]]
		
		matplotlib.pyplot.hist(obs_r2, log = True, bins = 50, range = (0,1), alpha = .5)
		matplotlib.pyplot.hist(exp_r2, log = True, bins = 50, range = (0,1), alpha =.5)
		pylab.xlim(0, 1)
		pylab.title("Blue = Observed, Green = Expected\n{}, Subset = {}".format(self.name, subset_key))
		pylab.ylabel("Count")
		pylab.xlabel("R2")
		
	# Observed vs Expected R2
	def plot_obs_exp_r2(self, subset_key = None):
		obs_r2 = self.r2_plink
		exp_r2 = self.lower_r2_estimate
		if subset_key is None:
			pass
		else:
			obs_r2 = obs_r2[subset_loci[subset_key]]
			exp_r2 = exp_r2[subset_loci[subset_key]]
			
		matplotlib.pyplot.hexbin(exp_r2, obs_r2,
			mincnt = 1, bins = 'log', gridsize = 30)
		pylab.ylim([-0.04, 1.04])
		pylab.xlim([-0.04, 1.04])
		pylab.title("exp R2 vs OBS R2\n{}, Subset = {}".format(self.name, subset_key))
		pylab.xlabel('Lower of [SF, WH] expected R2 ')
		pylab.ylabel('Observed R2')
		cb = matplotlib.pyplot.colorbar()
		cb.set_label('log(count)')
		matplotlib.pyplot.show()
	
	# Observed R2 vs Recombination Fraction
	def plot_c_obs_r2(self, subset_key = None):
		c_mat = self.c_mat
		obs_r2 = self.r2_plink
		if subset_key is None:
			pass
		else:
			c_mat = c_mat[subset_loci[subset_key]]
			obs_r2 = obs_r2[subset_loci[subset_key]]
		
		matplotlib.pyplot.hexbin(c_mat, obs_r2,
			mincnt = 1, bins = 'log', gridsize = 60,
			#xscale = 'log'
			 )
		pylab.ylim([-0.04, 1.04])
		pylab.xlim([-0.02, 0.52])
		pylab.title("c vs Observed R2\n{}, Subset = {}".format(self.name, subset_key))
		pylab.xlabel('Recombination Fraction (c)')
		pylab.ylabel('Observed R2')
		cb = matplotlib.pyplot.colorbar()
		cb.set_label('log(count)')
		matplotlib.pyplot.show()
		
	def plot_r2_hist(self, subset_key = None):
		obs_r2 = self.r2_plink
		if subset_key is None:
			pass
		else:
			obs_r2 = obs_r2[subset_loci[subset_key]]
		matplotlib.pyplot.hist(obs_r2, log = True)
		

		
# Calculate Summary Stats	
def read_R2_matrix(matrix_file):
	"""Can natively read *.gz files"""
	#R2 = numpy.loadtxt(fname = matrix_file)
	R2 = numpy.load(matrix_file)
	# limit values to 1
	R2[R2 > 1] = 1
	# collapse to 1D vector 
	R2 = R2[numpy.triu_indices_from(R2, k = 1)]
	return(R2)		
