import numpy 
import matplotlib.pyplot as pyplot
from scipy.stats import chi2


# This script will generate and plot the likelihoods for each chi2 distribution corresponding to an expected r2 value.  
# This allows calculation of the likelihood of observing a given r2 value under different recombination fractions.
# It is not yet clear how to generate a ML estimate of the c value given an observed r2 value,
# either within a single population, or across multiple populations.

# The chi2.fit([ .05], f0 = 1, floc = 0) function returns the MLE of the scale as equal to the data point for a single data point.
# for multiple data point it returns the mean of them.  i.e the mean r2 across populations = the MLE of the scale parameter.
# Too easy? - yes this doesn't hold the c value constant across populations (doh).  Each population has is own Ne, S, and therefore expected r2.

# Ne used in calculations of closesly linked loci should be historic Ne.


interval = .001
max = 10
left_intervals = numpy.arange(0, max, interval)
right_intervals = numpy.arange(interval, max + interval, interval) 
at_or_above_one = right_intervals >= 1.0
at_one = right_intervals == 1.0
def plot_likes(c_values, Ne, S):
    for c_value in c_values:
        r2_SF = expected_r2_SF(c_value, Ne, reshape = False) + 1./S
        r2_WH = expected_r2_WH(c_value, Ne, reshape = False) + 1./S
        r2 = min([r2_SF, r2_WH])
        interval_density = None
        count_loci = 1
        scaled_chi2_dist_frozen = chi2(1, scale = r2)
        scaled_left_cdf = scaled_chi2_dist_frozen.logcdf(left_intervals)
        scaled_right_cdf = scaled_chi2_dist_frozen.logcdf(right_intervals)
        scaled_interval_density = scaled_right_cdf - scaled_left_cdf
        scaled_interval_density *= count_loci
        #print(mean_r2, numpy.sum(scaled_interval_density), numpy.max(scaled_interval_density))
        if interval_density is None:
            interval_density = scaled_interval_density
        else:
            interval_density += scaled_interval_density       
        density_above_one = numpy.sum(interval_density[at_or_above_one])
        interval_density[at_or_above_one] = 0
        interval_density[at_one] = density_above_one
        pyplot.plot((right_intervals + left_intervals)/2., interval_density, linewidth= 1.0, alpha = .3, label = "{0:.4g}, {1:.3g}".format(c_value, r2))
    pyplot.xlim(-.04, 1.04)
    pyplot.ylim(.0000001, 1)
    pyplot.ylabel("Likelihood")
    pyplot.xlabel("Observed r^2")
    pyplot.semilogy()
    pyplot.legend(loc = "lower left", title = "c value | Expected r^2", fontsize = 8)
    pyplot.title("Ne = {}, S = {}".format(Ne, S))
        
        
def est_c_numerator(c):
    """
    (Matrix-aware)  
    Solves the expression:
    ((1-c)**2 + c**2)/(4*c - 2*c**2)
    for the supplied recombination fraction (c).
    numerator from  Weir & Hill 1981
    """
    # c = recombination fraction
    # return numerator from  Weir & Hill 1981
    # (r2 - (1/S)) * Ne = ((1-c)**2 + c**2)/(4*c - 2*c**2)
    #if c is 0:
    #   return(inf)
    #else:
    return (((1-c)**2 + c**2)/(4*c - 2*c**2))
      
def expected_r2_WH(c, Ne, bound = True, reshape = True):
    """
    Weir and Hill 1981?
    c - recombination fraction
    Ne - Effective Size
    S - Sample size
    And returns the projected r^2 value.
    """
    numerator = est_c_numerator(c)
    est_r2 = (numerator / Ne)
    if reshape == True:
        if c.__class__ == numpy.ndarray:
            numpy.fill_diagonal(est_r2, 1)
            if bound:
                est_r2[est_r2 > 1] = 1
    return est_r2         
    
def expected_r2_SF(c, Ne, bound = True, reshape = True):
    """
    Sved and Feldman (1973)
    c - recombination fraction
    Ne - Effective Size
    S - Sample size
    And returns the projected r^2 value.
    """
    est_r2 = 1./(1 + 2 * Ne * c *(2 - c))
    if reshape == True:
        if c.__class__ == numpy.ndarray:
            numpy.fill_diagonal(est_r2,1)
            if bound:
                est_r2[est_r2 > 1] = 1
    return est_r2 
        
        
possible_c_values = numpy.append(numpy.arange(0.0001, .02, .0005), numpy.arange(0.15, .55, .05))
#possible_r2 = expected_r2_WH(possible_c_values, 500, reshape = False)
        
        
plot_likes(possible_c_values, 2000, 50)
pyplot.xlim(-.04, 1.04)
pyplot.ylim(.0000001, 1)

my_chi2 = chi2(1, loc = 0)

chi2.fit([ .05], f0 = 1, floc = 0)
chi2.fit_loc_scale([ .4, .4], 1)

intervals = numpy.arange(0.01, 1, .01)
scales = numpy.arange(.1, 1, .1)

for sc in scales:
    dist = chi2(1, scale = sc)
    plot(intervals, dist.logpdf(intervals),alpha = .7, lw = 2, label = sc)
legend()

semilogy()
pyplot.ylim(.01, 10)


chi2.pdf(.3, 1, scale = .31)

plot