#!/usr/bin/env python
from __future__ import division
import simuOpt
import simuPOP 
import simuPOP.utils
import simuPOP.sampling

import os
import os.path
import itertools
import numpy
import zipfile
import datetime

from generate_genome import *
from simuPOP_functions import *

class simuPOP_batch:
	def __init__(self,
				name,
				# Constants
				ploidy 			= 2,
				map_function 	= "kosambi",
				mutation_rate 	= [],
				cM_unit 		= 10,
				# List variables
				replicates		= [],
				popsize 		= [],
				sampsize 		= [],
				generations 	= [],
				loci_num		= [],
				chr_num 		= [],
				chr_cM 			= [],
				scale_recombination = [1],	
				scale_mutation 		= [1]
				):
		# constants		
		self.name 			= name
		self.ploidy 		= ploidy 		
		self.map_function 	= map_function
		self.mutation_rate 	= mutation_rate
		self.cM_unit		= cM_unit
		# variables
		self.replicates 			= replicates
		self.popsize 				= popsize
		self.sampsize 				= sampsize
		self.loci_num				= loci_num
		self.chr_num 				= chr_num
		self.chr_cM 				= chr_cM
		self.scale_recombination 	= scale_recombination
		self.scale_mutation 		= scale_mutation
		self.generations			= generations
			
		
	def construct_parameters(self):
		param_combinations = itertools.product(self.replicates,
										self.popsize,
										self.loci_num,
										self.chr_num,
										self.chr_cM,
										self.scale_recombination,
										self.scale_mutation
										)
		param_keys = ['replicate','Ne','loci_num', 'chr_num', 'chr_cM', 'scale_recombination', 'scale_mutation']
		self.params = [dict(zip(param_keys, combination)) for combination in param_combinations]
		return(self.params)
	
	def run_all(self, out_dir, out_zip):
		batch_file_name = os.path.join(out_dir, self.name + ".batch")
		# set current date
		
		current_date = datetime.datetime.now().strftime("%c")
		with open(batch_file_name, 'w') as OUTFILE:
			OUTFILE.writelines([
				"{}\n".format(self.name),
				"{}\n".format(current_date),
				"target directory: {}\n".format(out_dir),
				"name: {}\n".format(self.name),
				"ploidy: {}\n".format(self.ploidy),
				"map_function: {}\n".format(self.map_function),
				"mutation_rate: {}\n".format(self.mutation_rate),
				"cM_unit: {}\n".format(self.cM_unit),
				"replicates: {}\n".format(self.replicates),
				"popsize: {}\n".format(self.popsize),
				"sampsize: {}\n".format(self.sampsize),
				"generations: {}\n".format(self.generations),
				"loci_num: {}\n".format(self.loci_num),
				"chr_num: {}\n".format(self.chr_num),
				"chr_cM: {}\n".format(self.chr_cM),
				"scale_recombination: {}\n".format(self.scale_recombination),
				"scale_mutation: {}\n".format(self.scale_mutation),
			])
			
			
		self.construct_parameters()
		for param_set in self.params:
			self._run_simuPOP(param_set = param_set, out_dir = out_dir, out_zip = out_zip)
	
	def _run_simuPOP(self, param_set, out_dir, out_zip):
		# unpack param_set
		print("param set:")
		print(param_set)
		REP = param_set['replicate']
		POPSIZE = param_set['Ne']
		LOCI_NUM = param_set['loci_num']
		CHR_NUM = param_set['chr_num']
		CHR_cM = param_set['chr_cM']
		SCALE_RECOMBINATION = param_set['scale_recombination']
		SCALE_MUTATION = param_set['scale_mutation']
		
		# Calculations needed
		if LOCI_NUM % CHR_NUM != 0:
			raise ValueError("Please make sure loci can be even allocated across chromosomes.")
		CHR_LOCI = int(LOCI_NUM / CHR_NUM)
		
		if self.generations == "equilibrium":
			#calculate an estimate of generations to equilibrium
			target_gen, eq_r2, method = generations_to_equilibrium_composite(Ne = POPSIZE, min_c = 1./(100 * self.cM_unit), max_gen = 1000, threshold= .0001, ini_r2 = 1.)
			# sample at 90%, 100%, and 110% of this estimate.
			print ("Target Generation: {}, Equilibrium r2: {}, method: {}, minimum c: {}".format(target_gen, eq_r2, method, 1./(100 * self.cM_unit)))
			#GENERATIONS_TO_SAMPLE = [int(target_gen* 0.9), target_gen, int(target_gen * 1.1)]
			GENERATIONS_TO_SAMPLE = [int(target_gen-2), int(target_gen-1),int(target_gen-0),int(target_gen+1),int(target_gen+2)]
			GENERATIONS_TO_SAMPLE += [x+20 for x in GENERATIONS_TO_SAMPLE]
			
		else:
			GENERATIONS_TO_SAMPLE = self.generations
		MAX_GENERATIONS = max(GENERATIONS_TO_SAMPLE)
		print("GENERATIONS_TO_SAMPLE: ")
		print(GENERATIONS_TO_SAMPLE)

		
		#[[x]* 30 for x in range(30)]
		LOCUS_NAMES = []
		iterit = itertools.count(1)
		for x in range(CHR_NUM):
			for y in range(CHR_LOCI):
				loc = next(iterit)
				LOCUS_NAMES.append(str(x+1)+"_"+str(loc))
		CHROM_NAMES = [str(100 + ii) for ii in range(CHR_NUM)]
		#print (CHROM_NAMES)


		
		# Generate Genome
		LOCI_POSITIONS, cM_DISTANCES = generate_genome_simuPOP(
										chrom_num = CHR_NUM, 
										chrom_cM = CHR_cM,
										chrom_loci = CHR_LOCI,
										map_function = self.map_function,
										unit = self.cM_unit
										)
		# Run simuPOP
		my_pop = simuPOP.Population(
							size = POPSIZE, # number of individuals
							ploidy = self.ploidy, 
							loci = [CHR_LOCI] * CHR_NUM, # number of loci per chromosome * number of chromosomes
							alleleNames = ('A','T'),
							lociPos = list(itertools.chain(*LOCI_POSITIONS)),# positions of the loci given cumulative, adjacent recombination fractions
							lociNames = LOCUS_NAMES,
							chromNames = CHROM_NAMES
							)
		my_pop.evolve(
			initOps = [
			  simuPOP.InitSex(),
			  simuPOP.InitGenotype(haplotypes = [[0,0],[0, 1], [1, 0], [1,1]], #possible haplotypes are [1,1,1,1,...] and [0,0,0,0,...]
				prop = [.25, .25, .25, .25]), # initial haplotype frequencies, exactly
			],
			matingScheme = simuPOP.RandomMating(
								# Recombination scaled to locus positions
								ops = simuPOP.Recombinator(intensity = 1 * SCALE_RECOMBINATION),
								# Ensures sex ratio = 1
								sexMode = (simuPOP.GLOBAL_SEQUENCE_OF_SEX, simuPOP.MALE, simuPOP.FEMALE)
												),
			preOps = [
				#Diallelic mutation model
				simuPOP.SNPMutator(u = self.mutation_rate * SCALE_MUTATION, v = self.mutation_rate * SCALE_MUTATION), 
			], 
			postOps=[
				simuPOP.PyOperator(func = process_subsample, 
											param = {
												'samp_sizes' : self.sampsize,
												'cM_unit' : self.cM_unit,
												'cM_distances': cM_DISTANCES,
												'intermediate_exp_r2': False,
												'genepop_out' : True,
												'map_ped_out' : True,
												'path' : out_dir,
												'popsize': POPSIZE,
												'loci_num': LOCI_NUM,
												'chr_num': CHR_NUM,
												'chr_cM' : CHR_cM,
												'scale_recombination' : SCALE_RECOMBINATION,
												'scale_mutation' : SCALE_MUTATION,
												'rep' : REP
											}, 
											at = GENERATIONS_TO_SAMPLE
								)
			],
			gen = MAX_GENERATIONS + 20
		  )