#!/usr/bin/env python
from __future__ import division
import random
import scipy.spatial.distance
import numpy
import os.path

from generate_genome import cM_to_c
from PARAMS import DEBUG

# TODO
    # Implement logging
    # output a 'genome' to a file



class Genome(object):
	def __init__(self, chr_num, chr_cM, chr_loci, cM_unit, map_function = "kosambi"):
		self.map_function = map_function
		self.cM_unit = cM_unit
		self.chr_num = chr_num
		self.chr_cM = chr_cM
		self.chr_loci = chr_loci
		if chr_cM > 0:
      		    self.cM_genome = self.generate_cM_genome(chr_num = self.chr_num, chr_cM = self.chr_cM, chr_loci = self.chr_loci, cM_unit = self.cM_unit)
      		    self.adjacent_c = self.generate_adjacent_c(self.cM_genome)
      		    self.c_matrix = self.genome_c_matrix(cM_genome = self.cM_genome, total_loci = int(chr_num * chr_loci), map_function = self.map_function)
      		    self.simuPOP_lociPos = self.set_simuPOP_lociPos()
      		else:
      		    print ("Creating Genome with free recombination")
      		    self.cM_genome = None
      		    self.adjacent_c = numpy.array(self.chr_num * [(self.chr_loci- 1) * [.5]])
      		    self.c_matrix = self.genome_c_matrix(cM_genome = self.cM_genome, total_loci = self.chr_num * self.chr_loci, map_function = self.map_function)
                    self.simuPOP_lociPos = self.set_simuPOP_lociPos()
                    
        def write_c_matrix_to_file(self, path, filename_base, npy_format = False):
            """Writes two files representing a complete genome.
                1) a binary file: "path/filename.npy"
                2) a gzipped text file: "path/filename.tsv.gz" 
            """
            if npy_format:
                binary_path = os.path.join(path, filename_base + '.npy')
                numpy.save(file = binary_path, arr = self.c_matrix)
            else:
                txt_path = os.path.join(path, filename_base + '.tsv.gz')
                #numpy.savetxt(fname = txt_path, X = self.c_matrix, delimiter = "\t")
                numpy.save(file = txt_path, arr = self.c_matrix)
        
	def generate_cM_chromosome(self, chr_cM, chr_loci, cM_unit):
		"""
		Returns a numpy.array of *num_loci* loci onto a chromosome *num_cM* units long.
		"""
		if chr_cM <= 0:
			raise Exception("Please select a positive cM as the chromosome length")
		else:
			positions = sorted(random.sample(range(int(chr_cM * cM_unit)), chr_loci))
			cM_chromosome = numpy.array([xx/float(cM_unit) for xx in positions])
			return (cM_chromosome)		
		
	def generate_cM_genome(self, chr_num, chr_cM, chr_loci, cM_unit):
		if chr_cM <= 0:
			raise Exception("Please select a positive cM as the chromosome length")
		if chr_num <= 0:
			raise Exception("Please select a positive number of chromosomes")
		cM_genome = None
		for each_chr in range(chr_num):
			cM_chrom =  self.generate_cM_chromosome(chr_cM, chr_loci, cM_unit)
			if cM_genome is None:
				cM_genome = cM_chrom.reshape(1, len(cM_chrom))
			else:
				cM_genome = numpy.vstack((cM_genome, cM_chrom))
		return(cM_genome)
                
	def generate_adjacent_c(self, cM_genome):
		adjacent_c_genome = None
		for cM_chrom in cM_genome:
			#locus_spacings = [(jj-ii) for ii, jj in zip(cM_chrom[:-1], cM_chrom[1:])]
			locus_spacings = numpy.diff(cM_chrom)
			adjacent_c_chrom = cM_to_c(locus_spacings, self.map_function)
			if adjacent_c_genome is None:
				adjacent_c_genome = adjacent_c_chrom.reshape(1, len(adjacent_c_chrom))
			else:
				adjacent_c_genome = numpy.vstack((adjacent_c_genome, adjacent_c_chrom))
		return (adjacent_c_genome)
			
			
	def chromosome_c_matrix(self, cM_chrom, map_function):
		"""
		Pairwise recombination fraction (c) calculated for a single linear chromosome.  
		Returns a square distance matrix
		"""
		cM_chrom = numpy.vstack(tuple(cM_chrom))
		square_cM_matrix = scipy.spatial.distance.cdist(cM_chrom, cM_chrom)
		square_c_matrix = cM_to_c(square_cM_matrix, map_function)
		return(square_c_matrix)

	
	def genome_c_matrix(self, cM_genome, total_loci, map_function):
		# you can get the adjacent c values with numpy.diag(full_c_matrix, k =1)
		if cM_genome is None:
			c_array = numpy.array([0.5 for xx in range(total_loci**2)])
			c_matrix = numpy.reshape(a = c_array, newshape = (total_loci, total_loci))
			return(c_matrix)
		else:
			total_loci = len(cM_genome.flatten())
			inital_c_array = numpy.array([0.5 for xx in range(total_loci**2)])
			full_c_matrix = numpy.reshape(a = inital_c_array, newshape = (total_loci, total_loci))
			start_pos = 0 
			last_pos = 0
			for each_chr in range(len(cM_genome)):
				#print(each_chr)
				partial_c_matrix = self.chromosome_c_matrix(cM_genome[each_chr], map_function = map_function)
				#print(partial_c_matrix)
				last_pos = start_pos + len(cM_genome[each_chr])
				full_c_matrix[start_pos:last_pos,start_pos:last_pos] = partial_c_matrix
				start_pos = last_pos
			return(full_c_matrix)
			
	def set_simuPOP_lociPos(self):
		cumsum_without_zeros = numpy.cumsum(self.adjacent_c, 1)
		initial_zeros = numpy.vstack([0]*len(self.adjacent_c))
		cumsum = numpy.hstack((initial_zeros, cumsum_without_zeros))
		return(list(cumsum.flatten()))
		
	def SIMCOAL_genome_strings(self, min_allele_freq):
		if DEBUG: print("min_allele_freq: {}".format(min_allele_freq))
		lines_for_SIMCOAL = []
		first_line = "//Number of chromosomes\n"
		second_line = "{} {}\n".format(self.chr_num, 1)
		lines_for_SIMCOAL.append(first_line)
		lines_for_SIMCOAL.append(second_line)
		# for each chromosome
		for xx in range(self.chr_num):
			first_line = "// Chromosome block structure\n"
			second_line = "{}\n".format(self.chr_loci)
			third_line = "//BLOCKS: data type, number of loci, recombination rate, min_allele_freq\n"
			lines_for_SIMCOAL.append(first_line)
			lines_for_SIMCOAL.append(second_line)
			lines_for_SIMCOAL.append(third_line)
			for c in self.adjacent_c[xx]:
				next_line = "SNP 1 {} {}\n".format(c, min_allele_freq)
				lines_for_SIMCOAL.append(next_line)
			#get last SNP as well, with no recombination after it
			next_line = "SNP 1 {} {}\n".format(0.0, min_allele_freq)
			lines_for_SIMCOAL.append(next_line)
		return(lines_for_SIMCOAL)

		
		

